class Runner:
    '''
    A Runner is an object that tells a :func:`.map_reduce` how to run on a local machine or cluster. Examples include
    :class:`.Local` and :class:`LocalMultiProc`.

    '''
    pass
